#include <iostream>
#include <random>
#include <sstream>
#include <string>
#include <vector>

#include "CartesianCell.h"
#include "CartesianGrid.h"

CartesianGrid::CartesianGrid(int rows, int cols)
    : mrows(rows),
      mcols(cols),
      mcells(rows, std::vector<CartesianCell*>(cols)) {
  // Initialize
  for (int i(0); i < rows; ++i) {
    for (int j(0); j < cols; ++j) {
      mcells[i][j] = new CartesianCell(i, j);
    }
  }
  // Set neighbors
  for (int i(0); i < rows; ++i) {
    for (int j(0); j < cols; ++j) {
      CartesianCell* cell = mcells[i][j];
      cell->north(getCell(i - 1, j));
      cell->east(getCell(i, j + 1));
      cell->south(getCell(i + 1, j));
      cell->west(getCell(i, j - 1));
    }
  }
  // Setup random
  std::mt19937 rng((std::random_device())());
  std::uniform_int_distribution<int> rangeRows(0, mrows - 1);
  std::uniform_int_distribution<int> rangeColumns(0, mcols - 1);
  mrandomRow = std::bind(rangeRows, rng);
  mrandomCol = std::bind(rangeColumns, rng);
}

CartesianGrid::~CartesianGrid() {
  foreachCell([](CartesianCell* cell) {
    if (cell != nullptr) delete cell;
  });
}

CartesianCell* CartesianGrid::getCell(int row, int col) const {
  if (row < 0 || row >= mrows) return nullptr;
  if (col < 0 || col >= mcols) return nullptr;
  return mcells[row][col];
}

CartesianCell* CartesianGrid::operator()(int row, int col) const {
  return getCell(row, col);
}

void CartesianGrid::foreachRow(
    std::function<void(std::vector<CartesianCell*>)> rowFunction) {
  auto begin = std::begin(mcells);
  auto end = std::end(mcells);
  std::for_each(begin, end, rowFunction);
}

void CartesianGrid::foreachCell(
    std::function<void(CartesianCell*)> cellFunction) {
  foreachRow([cellFunction](std::vector<CartesianCell*> row) {
    auto begin = std::begin(row);
    auto end = std::end(row);
    std::for_each(begin, end, cellFunction);
  });
}

void CartesianGrid::forRandomCell(
    std::function<void(CartesianCell*)> cellFunction) {
  int row(mrandomRow()), col(mrandomCol());
  cellFunction(getCell(row, col));
}

std::ostream& operator<<(std::ostream& out, const CartesianGrid& grid) {
  for (int i(0); i < grid.mrows; ++i) {
    std::vector<std::stringstream> stringRow(3);
    for (int j(0); j < grid.mcols; ++j) {
      CartesianCell* currentCell = grid.getCell(i, j);
      std::string north(currentCell->isLinked(currentCell->north()) ? "   "
                                                                    : "---");
      std::string south(currentCell->isLinked(currentCell->south()) ? "   "
                                                                    : "---");
      std::string west(currentCell->isLinked(currentCell->west()) ? " " : "|");
      std::string east(currentCell->isLinked(currentCell->east()) ? " " : "|");
      stringRow[0] << "+" << north << "+";
      stringRow[1] << west << "   " << east;
      stringRow[2] << "+" << south << "+";
    }
    out << stringRow[0].str() << std::endl;
    out << stringRow[1].str() << std::endl;
    out << stringRow[2].str() << std::endl;
  }
  return out;
}